﻿using System;
using System.Collections.Generic;
using System.Text;
using nicoAE.Plugin.Interface;
using nicoAE.Plugin.Utils;
using System.Drawing;
using System.Drawing.Imaging;
using nicoAE.Plugin.Controls;
using nicoAE.Plugin.Property;
using System.Runtime.InteropServices;
using nicoAE._2DUtils;
using nicoAE.Plugin.Threading;

namespace mes.Test
{
    public class WaveCurve : IEffect
    {
        IHost ho = null;

        #region IEffect メンバ

        public string PluginName
        {
            get
            {
                return "波形";
            }
        }

        public string Category
        {
            get
            {
                return "変形";
            }
        }

        public bool PresetProperty
        {
            get
            {
                return false;
            }
        }

        public MediaType CorrespondingType
        {
            get
            {
                return MediaType.Image | MediaType.Video;
            }
        }

        public void Initialize(IHost host, MediaInfo info)
        {
            ho = host;
        }

        public IPropertyEditControl[] GetControls()
        {
            return new IPropertyEditControl[] { new SelectableEditControl("波形"), new SingleEditControl("周期"), new SingleEditControl("振幅"), new RadianFEditControl("角度"), new RadianFEditControl("フェーズ") };
        }

        public IEffectProperty[] GetDefaultProperty()
        {
            return new IEffectProperty[] { new SelectableProperty("波形", new string[] { "サイン波", "矩形波", "三角波", "ノコギリ波" }, 0), new SingleProperty("周期", 2, float.MaxValue, 0), new SingleProperty("振幅", 30, float.MaxValue, 0), new RadianFProperty("角度", 0, 0), new RadianFProperty("フェーズ", 0, 0) };
        }

        public KeyFrame[] GetPresetProperty(int index, string propertyName)
        {
            throw new NotImplementedException();
        }

        public Bitmap ProcessingImage(Bitmap image, IEffectProperty[] property)
        {
            int wave = 0;
            float cycle = 0;
            float amp = 0;
            float angle = 0;
            float phase = 0;
            for (int i = 0; i < property.Length; i++)
            {
                switch (property[i].PropertyName)
                {
                    case "波形":
                        {
                            wave = ((SelectableProperty)property[i]).TypeSafeValue;
                            break;
                        }
                    case "周期":
                        {
                            cycle = ((SingleProperty)property[i]).TypeSafeValue;
                            break;
                        }
                    case "振幅":
                        {
                            amp = ((SingleProperty)property[i]).TypeSafeValue;
                            break;
                        }
                    case "角度":
                        {
                            angle = ((RadianFProperty)property[i]).Angle + 90;
                            break;
                        }
                    case "フェーズ":
                        {
                            phase = ((RadianFProperty)property[i]).Angle;
                            break;
                        }
                }
            }

            if (cycle != 0 && amp != 0)
            {
                int width = image.Width;
                int height = image.Height;
                byte[] data = new byte[width * height * 4];
                byte[] result = new byte[data.Length];
                BitmapData bd = image.LockBits(new Rectangle(new Point(), image.Size), ImageLockMode.ReadWrite, PixelFormat.Format32bppArgb);
                Marshal.Copy(bd.Scan0, data, 0, data.Length);

                //波形を掛ける方向を変えるのではなく、イメージを回転させて変形、そして元に戻す。
                Matrix3D matrix = new Matrix3D();
                Matrix3D imatrix = new Matrix3D();
                matrix.RotateAt(width / 2.0F, height / 2.0F, angle);
                imatrix.RotateAt(width / 2.0F, height / 2.0F, -angle);

                //変形順序:
                //1，回転(matrix.GetPoint(w, h))
                //2，縦方向に指定された波形に変形(amp * WaveGenerator(wave, sr - phase))
                //3，逆回転(imatrix.GetPoint(rp.X, rp.Y + sy))
                //4，とりあえずバイリニア補間。
                if (ho.MultiThread)
                {
                    //マルチスレッドによる処理。
                    Parallel.For(0, height, delegate(int h)
                    {
                        for (int w = 0; w < width; w++)
                        {
                            PointF rp = matrix.GetPoint(w, h);
                            double sr = rp.X * cycle * Math.PI / 180.0;
                            float sy = (float)(amp * WaveGenerator(wave, sr - phase));
                            int pos = (h * width + w) * 4;
                            byte[] dst = InterpolationMethods.Bilinear(data, imatrix.GetPoint(rp.X, rp.Y + sy), width, height);
                            result[pos] = dst[0];
                            result[pos + 1] = dst[1];
                            result[pos + 2] = dst[2];
                            result[pos + 3] = dst[3];
                        }
                    });
                }
                else
                {
                    //シングルスレッドによる処理。
                    for (int h = 0; h < height; h++)
                    {
                        for (int w = 0; w < width; w++)
                        {
                            PointF rp = matrix.GetPoint(w, h);
                            double sr = rp.X * cycle * Math.PI / 180.0;
                            float sy = (float)(amp * WaveGenerator(wave, sr - phase));
                            int pos = (h * width + w) * 4;
                            byte[] dst = InterpolationMethods.Bilinear(data, imatrix.GetPoint(rp.X, rp.Y + sy), width, height);
                            result[pos] = dst[0];
                            result[pos + 1] = dst[1];
                            result[pos + 2] = dst[2];
                            result[pos + 3] = dst[3];
                        }
                    }
                }

                Marshal.Copy(result, 0, bd.Scan0, result.Length);
                image.UnlockBits(bd);
            }
            return image;
        }

        public byte[] ProcessingAudio(byte[] audio, IEffectProperty[] property)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IDisposable メンバ

        public void Dispose() { }

        #endregion

        float WaveGenerator(int mode, double radian)
        {
            //Sin波以外は0～2πの範囲に丸める。
            if (mode != 0)
            {
                if (radian > 0)
                {
                    while (radian > Math.PI * 2.0)
                    {
                        radian -= Math.PI * 2.0;
                    }
                }
                else
                {
                    while (radian < 0.0)
                    {
                        radian += Math.PI * 2.0;
                    }
                }
            }
            //0:Sin波
            //1:矩形波
            //2:三角波
            //3:ノコギリ波
            switch (mode)
            {
                case 0:
                    {
                        return (float)Math.Sin(radian);
                    }
                case 1:
                    {
                        if (radian > Math.PI)
                        {
                            return -1.0F;
                        }
                        else
                        {
                            return 1.0F;
                        }
                    }
                case 2:
                    {
                        float inv = -1.0F;
                        if (radian > Math.PI)
                        {
                            radian -= Math.PI;
                            inv = 1.0F;
                        }
                        return (float)(radian / Math.PI * 2.0 - 1.0) * inv;
                    }
                case 3:
                    {
                        if (radian > Math.PI)
                        {
                            radian -= Math.PI;
                        }
                        return (float)(radian / Math.PI * 2.0 - 1.0);
                    }
                default:
                    {
                        return 0.0F;
                    }
            }
        }
    }
}
